#include <stdio.h>

// error checking macro
#define cudaCheckErrors(msg) \
    do { \
        cudaError_t __err = cudaGetLastError(); \
        if (__err != cudaSuccess) { \
            fprintf(stderr, "Fatal error: %s (%s at %s:%d)\n", \
                msg, cudaGetErrorString(__err), \
                __FILE__, __LINE__); \
            fprintf(stderr, "*** FAILED - ABORTING\n"); \
            exit(1); \
        } \
    } while (0)

const size_t N = 8ULL*1024ULL*1024ULL;  // data size
//const size_t N = 256*640; // data size
const int BLOCK_SIZE = 256;  // CUDA maximum is 1024

// naive atomic reduction kernel
__global__ 
void atomic_red(const float *gdata, float *out)
{
  size_t idx = threadIdx.x + blockDim.x * blockIdx.x;
  if (idx < N) atomicAdd(out, gdata[idx]);
}

__global__ 
void reduce(float *gdata, float *out)
{
    __shared__ float sdata[BLOCK_SIZE];

    int tid = threadIdx.x;
    sdata[tid] = 0.0f;
    size_t idx = threadIdx.x+blockDim.x*blockIdx.x;

    while (idx < N) {  // grid stride loop to load data
        sdata[tid] += gdata[idx];
        idx += gridDim.x*blockDim.x;  
    }

    for (unsigned int s = blockDim.x / 2; s > 0; s /= 2) {
        __syncthreads();
        if (tid < s) { 
            // parallel sweep reduction
            sdata[tid] += sdata[tid + s];
        }
        if (tid == 0) out[blockIdx.x] = sdata[0];
    }
}

 __global__ 
void reduce_a(float *gdata, float *out)
{
    __shared__ float sdata[BLOCK_SIZE];

    int tid = threadIdx.x;
    sdata[tid] = 0.0f;
    size_t idx = threadIdx.x + blockDim.x * blockIdx.x;

    while (idx < N) {  // grid stride loop to load data
        sdata[tid] += gdata[idx];
        idx += gridDim.x*blockDim.x;  
    }

    for (unsigned int s = blockDim.x / 2; s > 0; s /= 2) {
        __syncthreads();
        if (tid < s) {
            // parallel sweep reduction
            sdata[tid] += sdata[tid + s];
        }
    }
    
    if (tid == 0) atomicAdd(out, sdata[0]);
}

__global__ 
void reduce_ws(float *gdata, float *out)
{
     __shared__ float sdata[32];
     
     int tid = threadIdx.x;
     int idx = threadIdx.x + blockDim.x * blockIdx.x;
     
     float val = 0.0f;
     
     unsigned mask = 0xFFFFFFFFU;
     int lane = threadIdx.x % warpSize;
     int warpID = threadIdx.x / warpSize;
     
     while (idx < N) {  // grid stride loop to load 
         val += gdata[idx];
         idx += gridDim.x * blockDim.x;  
     }

     // First warp-shuffle reduction
     for (int offset = warpSize / 2; offset > 0; offset /= 2) {
         val += __shfl_down_sync(mask, val, offset);
     }
     
     if (lane == 0) sdata[warpID] = val;
   
     __syncthreads(); // put warp results in shared mem

    // Hereafter, just warp 0
    if (warpID == 0){
       // reload val from shared mem if warp existed
       val = (tid < blockDim.x / warpSize) ? sdata[lane] : 0;

       // Final warp-shuffle reduction
       for (int offset = warpSize/2; offset > 0; offset /= 2) {
           val += __shfl_down_sync(mask, val, offset);
       }

       if (tid == 0) atomicAdd(out, val);
   }
}

int main()
{
    float *h_A, *h_sum, *d_A, *d_sum;
    h_A = new float[N];  // allocate space for data in host memory

    h_sum = new float;
    for (int i = 0; i < N; i++)  // initialize matrix in host memory
        h_A[i] = 1.0f;
  
    cudaMalloc(&d_A, N*sizeof(float));  // allocate device space for A
    cudaMalloc(&d_sum, sizeof(float));  // allocate device space for sum
    cudaCheckErrors("cudaMalloc failure"); // error checking
  
    // copy matrix A to device:
    cudaMemcpy(d_A, h_A, N*sizeof(float), cudaMemcpyHostToDevice);
    cudaCheckErrors("cudaMemcpy H2D failure");
    cudaMemset(d_sum, 0, sizeof(float));
    cudaCheckErrors("cudaMemset failure");

    atomic_red<<<(N+BLOCK_SIZE-1)/BLOCK_SIZE, BLOCK_SIZE>>>(d_A, d_sum);
    cudaCheckErrors("atomic reduction kernel launch failure");

    // copy vector sums from device to host:
    cudaMemcpy(h_sum, d_sum, sizeof(float), cudaMemcpyDeviceToHost);
    cudaCheckErrors("atomic reduction kernel execution failure or cudaMemcpy H2D failure");
  
    if (*h_sum != (float)N) {
        printf("atomic sum reduction incorrect!\n"); 
        return -1;
    }
    printf("atomic sum reduction correct!\n");
  
    const int blocks = 640;
    cudaMemset(d_sum, 0, sizeof(float));
    cudaCheckErrors("cudaMemset failure");

    reduce_a<<<blocks, BLOCK_SIZE>>>(d_A, d_sum);
    cudaCheckErrors("reduction w/atomic kernel launch failure");

    // copy vector sums from device to host:
    cudaMemcpy(h_sum, d_sum, sizeof(float), cudaMemcpyDeviceToHost);
    cudaCheckErrors("reduction w/atomic kernel execution failure or cudaMemcpy H2D failure");
  
    if (*h_sum != (float)N) {
        printf("reduction w/atomic sum incorrect!\n"); 
        return -1;
    }
    printf("reduction w/atomic sum correct!\n");
  
    cudaMemset(d_sum, 0, sizeof(float));
    cudaCheckErrors("cudaMemset failure");

    reduce_ws<<<blocks, BLOCK_SIZE>>>(d_A, d_sum);
    cudaCheckErrors("reduction warp shuffle kernel launch failure");

    // copy vector sums from device to host:
    cudaMemcpy(h_sum, d_sum, sizeof(float), cudaMemcpyDeviceToHost);
    cudaCheckErrors("reduction warp shuffle kernel execution failure or cudaMemcpy H2D failure");
  
    if (*h_sum != (float)N) {
        printf("reduction warp shuffle sum incorrect!\n"); 
        return -1;
    }
    printf("reduction warp shuffle sum correct!\n");
  
    return 0;
}
